#include <iostream>
#include <cmath>
#include "logfunc.h"
#include "logstuff.h"

using namespace std;

typedef double prec;
const int max_n = 35;
logfactorial<prec, max_n> logfac;


template<int c>
int shift(int seq)
{
  if(seq & 1)
    return (seq >> 1) ^ ((1 << c - 1) - 1);
  else
    return seq >> 1;
}

template<int c>
int n_seqgates(int i_seq0, int i_seq1, int o_seq)
{
  o_seq = shift<c>(o_seq);
  return 
    (o_seq == 0) + 
    (o_seq == i_seq0) + 
    (o_seq == i_seq1) +
    (o_seq == (i_seq0 ^ i_seq1)) +
    (o_seq == (i_seq0 & i_seq1)) +
    (o_seq == (i_seq0 | i_seq1)) +
    (o_seq == (~i_seq0 & i_seq1)) +
    (o_seq == (i_seq0 & ~i_seq1));
}

string pname(int n)
{
  const char *v[] = {"a", "b0", "b1", "c0", "b2", "d", "c1", "b3"};
  return string(v[n]);
}

double etest(int a, int b0, int b1, int b2, int b3,
	int c0, int c1, int d)
{
	int n_seq[8];
	n_seq[0] = a;
	n_seq[1] = b0;
	n_seq[2] = b1;
	n_seq[3] = c0;
	n_seq[4] = b2;
	n_seq[5] = d;
	n_seq[6] = c1;
	n_seq[7] = b3;
	double entropy = 0;

	for(int i = 0; i < 8; i++)
		if(n_seq[i])
		{
			double f = 0.;
			for(int l = 0; l < 8; l++)
				if(n_seq[l])
					f += n_seqgates<4>(l,l,i)*n_seq[l]*n_seq[l];
			for(int j = 0; j < 8; j++)
				if(n_seq[j])
					for(int l = 0; l < j; l++)
						if(n_seq[l])
							f += 2*n_seqgates<4>(j,l,i)*n_seq[j]*n_seq[l];
//printf("tt(%s): %d * %f\n", pname(i).c_str(), n_seq[i], f/2);
			entropy += LOGFUNC(f/2)*n_seq[i];
			entropy -= logfac(n_seq[i]);
		}

  return entropy;
}


prec fourcyc(int n)
{
	prec sum = -100;
//int terms = 0;
	for(int a = 0; a < n; a++)
	{
		int n_a = n + a;
		for(int d = n - a - 1; d >= 0; d--)
		{
			int nleft = n - a - d;
			for(int c0 = nleft; c0 >= 0; c0--)
			{
				int ar_t = 2*(2*n*a + c0*c0 + d*d);
				for(int c1 = nleft - c0; c1 >= 0; c1--)
				{
					int c = c0 + c1;
					int b = nleft - c0 - c1;
					int cd_c0c1 = c*d + c0*c1;
					int ar_t2 = ar_t + b*b + 2*(c1*c1 + b*(c+d)) + cd_c0c1;
					int tmp = 4*a + 2*b;
					int c0r_t = c1 * (tmp + 2*c1) + cd_c0c1;
					int c1r_t = c0 * (tmp + 2*c0) + cd_c0c1;
					int dr_t = d * (tmp + 2*d) + cd_c0c1;
					for(int b0 = b; b0 >= 0; b0--)
					{
						for(int b1 = b - b0; b1 >= 0; b1--)
						{
							int ar_t3 = ar_t2 + b0*b0 + b1*b1;
							int b2_b3 = b - b0 - b1;
							for(int b2 = b2_b3; b2 >= 0; b2--)
							{
								int b3 = b2_b3 - b2;
								int ar = ar_t3 + b2*b2 + b3*b3;
								if(a && !ar) continue;
								int dr = dr_t + 2*(b0*b2 + b1*b3);
								if(d && !dr) continue;
								int c0r = c0r_t + 2*(b0*b3 + b1*b2);
								if(c0 && !c0r) continue;
								int c1r = c1r_t + 2*(b0*b1 + b2*b3);
								if(c1 && !c1r) continue;
								int b0r = 2*(b3*n_a + b0*c1 + b1*d + b2*c0) + cd_c0c1;
								if(b0 && !b0r) continue;
								int b1r = 2*(b0*n_a + b1*c0 + b2*d + b3*c1) + cd_c0c1;
								if(b1 && !b1r) continue;
								int b2r = 2*(b1*n_a + b2*c1 + b3*d + b0*c0) + cd_c0c1;
								if(b2 && !b2r) continue;
								int b3r = 2*(b2*n_a + b3*c0 + b0*d + b1*c1) + cd_c0c1;
								if(b3 && !b3r) continue;
								prec t = 0;
								if(a) t += a*LOGFUNC(prec(ar)) - logfac(a);
								if(b0) t += b0*LOGFUNC(prec(b0r)) - logfac(b0);
								if(b1) t += b1*LOGFUNC(prec(b1r)) - logfac(b1);
								if(b2) t += b2*LOGFUNC(prec(b2r)) - logfac(b2);
								if(b3) t += b3*LOGFUNC(prec(b3r)) - logfac(b3);
								if(c0) t += c0*LOGFUNC(prec(c0r)) - logfac(c0);
								if(c1) t += c1*LOGFUNC(prec(c1r)) - logfac(c1);
								if(d) t += d*LOGFUNC(prec(dr)) - logfac(d);
								sum = ln_add(sum, t);
							}
						}
					}
				}
			}
		}
	}
//printf("terms: %d\n", terms);
	if(sum < 0) return 0.0;
	sum += logfac(n) - prec(n) * (2. * LOGFUNC(prec(n)) + 2 * M_LN2);
	return exp(sum) / 4.0;
}

int main()
{
	printf("#%s\t%s\t%s\t%s\n", "n", "4c_nodes", "estim.", "rel_err");
	for(int n = 1; n < max_n; n += 1 + (n >> 3))
	{
		prec cyc = fourcyc(n);
		prec est = prec(n) / 30.;
		printf("%d\t%f\t%f\t%f\n", n, cyc, est, est / cyc - 1.);
		fflush(stdout);
	}
	return 0;
}

