// Discards degenerate minima from folding of corners and rotation of a side.
// Reads possmin structs with SAW indexes according to sa9n??.out.

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <stdint.h>
#include <cstring>

const int NNODMAX = 34;
const int STARTP = NNODMAX, ARRW = NNODMAX*2+1;
const char INFSTR[] = "qqqn??.q",  SATABFSTR[] = "sa9n??.out";
const char TEXTFSTR[] = "HPn??.txt", OUTFSTR[] = "nodeg4n??.out";

short nnod, maxconns;
int nrarr[ARRW][ARRW];
short dseq[NNODMAX];  //directions 1=U, 2=D, 3=R, 4=L; nnod-1 of these
int ofs = 1;

struct possmin
{
  int32_t index;
  int32_t Hseq;
};

void initmaxconns(void)
{
   int sqrt4nnod;
   for(sqrt4nnod = 1; sqrt4nnod*sqrt4nnod < 4*nnod; sqrt4nnod++);
   maxconns = nnod - sqrt4nnod + 1;
}

int nHneib(int x, int y, int Hseq)
{
  int t, nn = 0;
  if((t=nrarr[x+1][y]-ofs) >= 0 && (Hseq >> t)&1) nn++;
  if((t=nrarr[x-1][y]-ofs) >= 0 && (Hseq >> t)&1) nn++;
  if((t=nrarr[x][y+1]-ofs) >= 0 && (Hseq >> t)&1) nn++;
  if((t=nrarr[x][y-1]-ofs) >= 0 && (Hseq >> t)&1) nn++;
  return nn;
}

short movepos(int &x, int &y, int d)
{
   switch (d) {
     case 1: x++; return 1;
     case 2: x--; return 1;
     case 3: y++; return 1;
     case 4: y--; return 1;
   }
   return 0;
}

short movepos(int &x, int &y)
{
   switch (dseq[nrarr[x][y]-ofs]) {
     case 1: x++; return 1;
     case 2: x--; return 1;
     case 3: y++; return 1;
     case 4: y--; return 1;
   }
   return 0;
}

short readseq1(int& deg, std::ifstream &inf)
{
  static uint16_t pkseq[4];
  int i;

  inf.read((char *)pkseq,8);
  if(inf.eof()) return 0;

  dseq[0]=1;
  for(i=0; i<nnod-2; i++)
    dseq[i+1] = (( pkseq[i>>3] >> ((i & 7) << 1) ) & 3) + 1;

  deg = pkseq[3] >> 15;

  return 1;
}

void setupnrarr()
{
  int x=STARTP, y=STARTP, nr;
 
  ofs += nnod;
  if(ofs < 0) {std::cerr << "offset negative in setupnrarr\n"; exit(0);}

  nrarr[x][y] = ofs;
  x++;
  nrarr[x][y] = ofs + 1;

  for(nr=2; nr<nnod; nr++) {
    if(!movepos(x,y)){ std::cerr << "Bug 1 in setupnrarr\n"; exit(0); }
    nrarr[x][y] = ofs + nr;
  }
}

inline int isH(int nr, int Hseq)
{
  return (Hseq >> nr) & 1;
}

int lnotnegnr(int tnr, int nr)
{
  return tnr >= 0 && tnr < nr;
}

int hHbnd(int tnr, int nr, int Hseq)
{
  return tnr > nr+1 && isH(tnr, Hseq);
}

int lHbnd(int tnr, int nr, int Hseq)
{
  return tnr >= 0 && tnr < nr-1 && isH(tnr, Hseq);
}

int issfdeg(int Hseq)
{
  int x, y;
  int nbrb = 0;  //# of broken interactions
  int nd13, nd23, nd24, nd14; //# of overlapping ( =0 for possible locking) for 4 diag. moved
  int nb13, nb23, nb24, nb14; //# of new interactions
  nd13 = nd23 = nd24 = nd14 = 0;
  nb13 = nb23 = nb24 = nb14 = 0;
  x = y = STARTP;

  for(int nr = 0; nr < nnod; nr++, movepos(x,y)) {
    if(isH(nr, Hseq)){
      //new broken interactions
      if(hHbnd(nrarr[x+1][y]-ofs, nr, Hseq)) nbrb++;
      if(hHbnd(nrarr[x-1][y]-ofs, nr, Hseq)) nbrb++;
      if(hHbnd(nrarr[x][y+1]-ofs, nr, Hseq)) nbrb++;
      if(hHbnd(nrarr[x][y-1]-ofs, nr, Hseq)) nbrb++;
      //ex-broken interactions
      if(lHbnd(nrarr[x+1][y]-ofs, nr, Hseq)) nbrb--;
      if(lHbnd(nrarr[x-1][y]-ofs, nr, Hseq)) nbrb--;
      if(lHbnd(nrarr[x][y+1]-ofs, nr, Hseq)) nbrb--;
      if(lHbnd(nrarr[x][y-1]-ofs, nr, Hseq)) nbrb--;
      if(nbrb < 0) {std::cerr << "Bug 1 in sfpossmin\n"; exit(0); }
    }
    	
    //will be on later
    if(nrarr[x+1][y+1]-ofs > nr) nd13++;
    if(nrarr[x-1][y+1]-ofs > nr) nd23++;
    if(nrarr[x-1][y-1]-ofs > nr) nd24++;
    if(nrarr[x+1][y-1]-ofs > nr) nd14++;
    //were on earlier
    if(lnotnegnr(nrarr[x-1][y-1]-ofs, nr)) nd13--;
    if(lnotnegnr(nrarr[x+1][y-1]-ofs, nr)) nd23--;
    if(lnotnegnr(nrarr[x+1][y+1]-ofs, nr)) nd24--;
    if(lnotnegnr(nrarr[x-1][y+1]-ofs, nr)) nd14--;
    if(nd13<0 || nd23<0 || nd24<0 || nd14<0) {std::cerr << "Bug 2 in sfpossmin\n"; exit(0);}

    if(isH(nr, Hseq)){
      //will bind later
      if(hHbnd(nrarr[x+1][y]-ofs, nr, Hseq)) {nb13++; nb14++;}
      if(hHbnd(nrarr[x-1][y]-ofs, nr, Hseq)) {nb23++; nb24++;}
      if(hHbnd(nrarr[x][y+1]-ofs, nr, Hseq)) {nb13++; nb23++;}
      if(hHbnd(nrarr[x][y-1]-ofs, nr, Hseq)) {nb14++; nb24++;}
      if(hHbnd(nrarr[x+2][y+1]-ofs, nr, Hseq)) nb13++;
      if(hHbnd(nrarr[x+1][y+2]-ofs, nr, Hseq)) nb13++;
      if(hHbnd(nrarr[x-2][y+1]-ofs, nr, Hseq)) nb23++;
      if(hHbnd(nrarr[x-1][y+2]-ofs, nr, Hseq)) nb23++;
      if(hHbnd(nrarr[x-2][y-1]-ofs, nr, Hseq)) nb24++;
      if(hHbnd(nrarr[x-1][y-2]-ofs, nr, Hseq)) nb24++;
      if(hHbnd(nrarr[x+2][y-1]-ofs, nr, Hseq)) nb14++;
      if(hHbnd(nrarr[x+1][y-2]-ofs, nr, Hseq)) nb14++;
      //were bound earlier
      if(lHbnd(nrarr[x-1][y]-ofs, nr, Hseq)) {nb13--; nb14--;}
      if(lHbnd(nrarr[x+1][y]-ofs, nr, Hseq)) {nb23--; nb24--;}
      if(lHbnd(nrarr[x][y-1]-ofs, nr, Hseq)) {nb13--; nb23--;}
      if(lHbnd(nrarr[x][y+1]-ofs, nr, Hseq)) {nb14--; nb24--;}
      if(lHbnd(nrarr[x-2][y-1]-ofs, nr, Hseq)) nb13--;
      if(lHbnd(nrarr[x-1][y-2]-ofs, nr, Hseq)) nb13--;
      if(lHbnd(nrarr[x+2][y-1]-ofs, nr, Hseq)) nb23--;
      if(lHbnd(nrarr[x+1][y-2]-ofs, nr, Hseq)) nb23--;
      if(lHbnd(nrarr[x+2][y+1]-ofs, nr, Hseq)) nb24--;
      if(lHbnd(nrarr[x+1][y+2]-ofs, nr, Hseq)) nb24--;
      if(lHbnd(nrarr[x-2][y+1]-ofs, nr, Hseq)) nb14--;
      if(lHbnd(nrarr[x-1][y+2]-ofs, nr, Hseq)) nb14--;
      if(nb13<0 || nb23<0 || nb24<0 || nb14<0) {std::cerr << "Bug 3 in sfpossmin\n"; exit(0);}
    }

    /*std::cout << x-STARTP << ", " << y-STARTP << ": " << dseq[nr] <<"\n";
    std::cout << "nd      " <<" "<< nd13 <<" "<< nd23 <<" "<< nd24 <<" "<< nd14 <<"\n";
    std::cout << "nb      " <<" "<< nb13 <<" "<< nb23 <<" "<< nb24 <<" "<< nb14 <<"\n";
    std::cout << "nb-nbrb " <<" "<< nb13-nbrb <<" "<< nb23-nbrb <<" "<< nb24-nbrb <<" "<< nb14-nbrb <<"\n";*/
    	
    switch(dseq[nr]) {
      case 1: if((nd13 == 0 && nb13 >= nbrb) || (nd14 == 0 && nb14 >= nbrb)) return 1; break;
      case 2: if((nd23 == 0 && nb23 >= nbrb) || (nd24 == 0 && nb24 >= nbrb)) return 1; break;
      case 3: if((nd13 == 0 && nb13 >= nbrb) || (nd23 == 0 && nb23 >= nbrb)) return 1; break;
      case 4: if((nd14 == 0 && nb14 >= nbrb) || (nd24 == 0 && nb24 >= nbrb)) return 1; break;
    }
  }
  if(nbrb) {std::cerr << "Bug 4 in sfpossmin\n"; exit(0); }
  if(nd13 || nd23 || nd24 || nd14) {std::cerr << "Bug 5 in sfpossmin\n"; exit(0);}
  if(nb13 || nb23 || nb24 || nb14) {std::cerr << "Bug 6 in sfpossmin\n"; exit(0);}

  return 0;
}
int isdfdeg(int Hseq)
{
  int x1,y1,x2,y2,x3,y3,x2p,y2p,nr3;
  x1 = y1 = STARTP;
  x2 = y2 = STARTP;
  x2++;
  x3 = x2;
  y3 = y2;

  for(nr3=2; nr3 < nnod; nr3++){
    if(!movepos(x3,y3)) {std::cerr << "bug 1 in isdfdeg\n"; exit(0);}
    x2p = x1 + x3 - x2;
    if(x2p != x2){
      y2p = y1 + y3 - y2;
      if(nrarr[x2p][y2p] < ofs && (!isH(nr3-1, Hseq) || nHneib(x2p,y2p,Hseq) >= nHneib(x2,y2,Hseq))) return 1;
    }
    x1 = x2; y1 = y2;
    x2 = x3; y2 = y3;
  }

  return 0;
}

void printdseq()
{
  for(int i=0; i<nnod-1; i++)
    std::cout << dseq[i] << " ";
  std::cout << "\n";
}

void strnum(char str[], int nr)
{
  int i;
  for(i=0; str[i] != '?'; i++);
  str[i++] = '0' + nr / 10;
  str[i] = '0' + nr % 10;
}

void writesaw(std::ofstream &textf)
{
  for(int i=0; i<nnod-1; i++)
    switch(dseq[i]) {
      case 1: textf << 'U'; break;
      case 2: textf << 'D'; break;
      case 3: textf << 'R'; break;
      case 4: textf << 'L'; break;
    }
  textf << '\n';
}

int nbonds(int Hseq)
{
  int x = STARTP, y = STARTP;
  int nbnd = 0;

  for(int nr = 0; nr < nnod; nr++, movepos(x,y))
    if(isH(nr, Hseq)){
      if(lHbnd(nrarr[x+1][y]-ofs, nr, Hseq)) nbnd++;
      if(lHbnd(nrarr[x-1][y]-ofs, nr, Hseq)) nbnd++;
      if(lHbnd(nrarr[x][y+1]-ofs, nr, Hseq)) nbnd++;
      if(lHbnd(nrarr[x][y-1]-ofs, nr, Hseq)) nbnd++;
    }

  return nbnd;
}

void writeHPseq(int Hseq, std::ofstream &textf)
{
  textf << "  ";
  for(int i=0; i<nnod; i++)
    if(isH(i, Hseq))
      textf << 'H';
    else
      textf << 'P';
  textf << ' ' << nbonds(Hseq) << '\n';
}

int main(int argc, char **argv)
{
  int i,j, isdeg=0, nnotdeg=0, nnotHseqdeg=0;
  char infstr[16], satabfstr[16], outfstr[16], textfstr[16];
  possmin pmin;

  for(i=0; i<ARRW; i++)
    for(j=0; j<ARRW; j++)
      nrarr[i][j]=0;

/*  std::cout << "n = ";
  cin >> nnod;*/
	if(argc != 2)
	{
		std::cerr << "Usage: "<<argv[0]<<" <N>\n";
		exit(1);
	}
	nnod = atoi(argv[1]);

  std::strcpy(infstr, INFSTR);
  strnum(infstr, nnod);
  std::ifstream inf(infstr, std::ios::in|std::ios::binary);
  if(!inf){std::cerr << "Unable to open " << infstr << '\n'; exit(0);}

  std::strcpy(satabfstr, SATABFSTR);
  strnum(satabfstr, nnod);
  std::ifstream satabf(satabfstr, std::ios::in|std::ios::binary);
  if(!satabf){std::cerr << "Unable to open " << satabfstr << '\n'; exit(0);}

  std::strcpy(outfstr, OUTFSTR);
  strnum(outfstr, nnod);
  std::ofstream outf(outfstr, std::ios::out|std::ios::trunc);
  if(!outf){std::cerr << "Unable to create " << outfstr << '\n'; exit(0);}

  std::strcpy(textfstr, TEXTFSTR);
  strnum(textfstr, nnod);
  std::ofstream textf(textfstr, std::ios::out|std::ios::trunc);
  if(!textf){std::cerr << "Unable to create " << textfstr << '\n'; exit(0);}

  std::cout << infstr << " -> " << outfstr << ", " << textfstr << "\n";

  int sawindex = -1, nwrittensaw = 0;
  int sawwritten = 0;
  while(inf.read((char *)&pmin, sizeof(possmin))){
    while(sawindex < pmin.index) {
      sawindex++;
      readseq1(isdeg, satabf);
      sawwritten = 0;
    }
    if(sawindex > pmin.index) {std::cerr << "Entries in " << infstr << " must be sorted be index (asc)"; exit(0);}
    if(!isdeg){
      setupnrarr();
      if(!isdfdeg(pmin.Hseq) && !issfdeg(pmin.Hseq)) {
        //printdseq();
        if(!sawwritten) {
          writesaw(textf);
          sawwritten = 1;
          nwrittensaw++;
        }
        writeHPseq(pmin.Hseq, textf);
        outf.write((char *)&pmin, sizeof(possmin));

        nnotHseqdeg++;
      }
      nnotdeg++;
    }
  }

  inf.close();
  satabf.close();
  outf.close();
  textf.close();

  std::cout << nnotHseqdeg << " HP minima with " << nwrittensaw << " different structures\n";
}

