/*
	Copyright Carl Troein 2000-2004

	Syntax: sort_p2_input [input_file output_file]
	Default values: "n_XX.conn", "n_XX.conn2"

*/

#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <map>
#include "constants.h"

#define MAXTEMPFILES 200
#define MAXFILESIZE 16777216
#define MAXFILEENTRIES (MAXFILESIZE / sizeof(struct node_d))

#define TEMPINFILENAME "infile.temp"

FILE *infile, *outfile;
FILE *tempfiles[MAXTEMPFILES];
int tempfileentries[MAXTEMPFILES];
std::map<int,int> entrycount;
int nr_tempfiles;
struct node_d *d_arr;

// For the subset thing
struct forkend
{
	unsigned int number;			// The highest bit is a flag, degen.
	unsigned short *subsets;	// The subsets, if any. subsets[0] is the number of subsets.
};
struct connfork
{
	union h
	{
		struct forkend *fe;		// if isend
		struct connfork *cf;		// if not isend
	} has;
	struct connfork *hasnot;
	unsigned char conn;
	unsigned char isend;
};
struct connfork *firstfork = NULL;

// For the saving of stuff.
int cursetnumber = 0;
int connlencount[3][MAXNRCONNS + 1];	// Total, degen, subsets.


void alloc_and_clear(char *outfn)
{
	outfile = fopen(outfn, "wb");
	if(!outfile)
	{
		perror("Can't open output file");
		exit(1);
	}

	d_arr = (struct node_d *)malloc(sizeof(struct node_d) * MAXFILEENTRIES);
	if(!d_arr)
	{
		printf("Out of memory!\n");
		exit(1);
	}
	memset(tempfileentries, 0, sizeof(tempfileentries));
	memset(connlencount, 0, sizeof(connlencount));
}

// My own memory allocation without a free()
void *mymalloc(int len)
{
	static unsigned long pointer;
	static int bytesleft = 0;
	void *poin;

	if(len > bytesleft)
	{
		pointer = (unsigned long)malloc(65536);
		if(!pointer)
		{
			printf("Out of bloody memory! Buy some more!\n");
			exit(1);
		}
		bytesleft = 65536;
	}
	bytesleft -= len;
	poin = (void *)pointer;
	pointer += len;
	memset(poin, 0, len);
	return poin;
}

int entrynum(int len, int ch0, int ch1)
{
	// We want the longest first, and must therefore flip the length.
	return ((MAXNRCONNS - len) * MAXCONNR + ch0) * MAXCONNR + ch1;
}

int cmpconn(unsigned char *a, unsigned char *b)
{
	if(*a < *b) return -1;
	return 1;
}
// Set the unused part of part2_input->conns to all 0 and sort the connections.
// Returns the length of the thingy.
inline int safenullify(unsigned char *conns)
{
	int j, len = -1;
	for(j = 0; j < MAXNRCONNS; j++)
	{
		if(len >= 0) conns[j] = 0;
		else if(!conns[j])
		{
			if(j > 1) qsort(conns, j, 1, (int (*)(const void *,const void *))cmpconn);
			len = j;
		}
	}
	if(len < 0)
	{
		qsort(conns, MAXNRCONNS, 1, (int (*)(const void *,const void *))cmpconn);
		return MAXNRCONNS;
	}
	return len;
}

void create_temp_input(char *from)
{
	FILE *myinfile;
	int entries, curentry, offset;
	struct part2_input *p_arr;

	p_arr = (struct part2_input *)d_arr;
	myinfile = fopen(from, "rb");
	if(!myinfile)
	{
		perror("Can't open input file");
		exit(1);
	}
	infile = fopen(TEMPINFILENAME, "wb+");
	if(!infile)
	{
		perror("Can't open temporary input file");
		exit(1);
	}
	offset = 0;
	while(1)
	{
		entries = fread(p_arr, sizeof(struct part2_input), MAXFILEENTRIES, myinfile);
		if(entries < 1) break;
		for(curentry = entries - 1; curentry >= 0; curentry--)
		{
			memmove(&d_arr[curentry].conns, &p_arr[curentry].conns, sizeof(struct part2_input));
			d_arr[curentry].number = curentry + offset;
			d_arr[curentry].wilds = 0;
			if(p_arr[curentry].degen) d_arr[curentry].number |= 0x80000000;
			d_arr[curentry].connlen = safenullify(d_arr[curentry].conns);
			entrycount[entrynum(d_arr[curentry].connlen, d_arr[curentry].conns[0],
				d_arr[curentry].conns[1])]++;
		}
		fwrite(d_arr, sizeof(struct node_d), entries, infile);
		offset += entries;
	}
	printf("Loaded %d entries to temporary input file.\n", offset);
	fclose(myinfile);
	rewind(infile);
	// We might want to delete the original input file now.
//	remove(from);
}

void calc_and_open_tempfiles(void)
{
	int entries, curfile;
	char namebuf[40];

	entries = 0;
	curfile = 0;
	nr_tempfiles = 0;
	for(std::map<int,int>::iterator iter = entrycount.begin(); iter != entrycount.end(); iter++)
	{
		if(entries + (*iter).second > (int)MAXFILEENTRIES)
		{
			// This file is full. Skip to the next or choke.
			if(entries == 0)	// One bunch is TOO big. Choke!
			{
				printf("Error: A bunch is too big. %d > %zd\n", (*iter).second,
					MAXFILEENTRIES);
				printf("You need to raise MAXFILESIZE to at least %zd\n",
					((*iter).second * sizeof(struct node_d)));
				exit(1);
			}
			entries = 0;
			if(++curfile == MAXTEMPFILES)
			{
				printf("Error: Temporary file count exceeded.\n");
				printf("Raise MAXFILESIZE or MAXTEMPFILES.\n");
				exit(1);
			}
		}
		entries += (*iter).second;
		tempfileentries[curfile] += (*iter).second;
		(*iter).second = curfile;
	}
	nr_tempfiles = curfile + 1;
	for(; curfile >= 0; curfile--)
	{
		sprintf(namebuf, "tmp_%d.out", curfile);
		tempfiles[curfile] = fopen(namebuf, "wb+");
		if(!tempfiles[curfile])
		{
			perror("Can't open temporary file");
			exit(1);
		}
	}
}

void distribute_temp_input(void)
{
	int entries, curentry, fnum, offset;

	offset = 0;
	while(1)
	{
		entries = fread(d_arr, sizeof(struct node_d), MAXFILEENTRIES, infile);
		if(entries < 1) break;
		for(curentry = entries - 1; curentry >= 0; curentry--)
		{
			fnum = entrycount[entrynum(d_arr[curentry].connlen, d_arr[curentry].conns[0],
				d_arr[curentry].conns[1])];
			fwrite(&d_arr[curentry], sizeof(struct node_d), 1, tempfiles[fnum]);
		}
		offset += entries;
	}
	printf("Distributed entries to %d temporary files.\n", nr_tempfiles);
	fclose(infile);
	remove(TEMPINFILENAME);
	for(fnum = 0; fnum < nr_tempfiles; fnum++)
	{
		rewind(tempfiles[fnum]);
	}
}

int comparenodes(struct node_d *n1, struct node_d *n2)
{
	if(n1->connlen == n2->connlen)
	{
		return memcmp(n1->conns, n2->conns, n1->connlen);
	}
	return (n1->connlen > n2->connlen) ? -1 : 1;
}

void quick_sort(int length)
{
	qsort(d_arr, length, sizeof(struct node_d),
		(int (*)(const void *,const void *))comparenodes);
}

// Generates a subset mask.
unsigned short subsetmask(unsigned char *s, unsigned char *subs, int sulen)
{
	unsigned short mask = 0;
	int bitpos = 0;
	while(1)
	{
		if(*s == *subs)
		{
			mask |= (1 << bitpos);
			subs++;
			if(--sulen == 0) return mask;
		}
		s++;
		bitpos++;
	}
}

void try_add_subset(struct node_d *node)
{
	unsigned char *conns = node->conns;
	int len = node->connlen;

	// One special case:
	if(!firstfork)
	{
		struct connfork *fork = NULL, *lfork;
		firstfork = (struct connfork *)mymalloc(sizeof(struct connfork));
		firstfork->conn = conns[0];
		if(len > 1)
		{
			lfork = firstfork;
			for(int i = 1; i < len; i++)
			{
				fork = (struct connfork *)mymalloc(sizeof(struct connfork));
				fork->conn = conns[i];
				lfork->has.cf = fork;
				lfork = fork;
			}
		}
		else fork = firstfork;
		fork->isend = 1;
		fork->has.fe = (struct forkend *)mymalloc(sizeof(struct forkend));
		fork->has.fe->subsets = NULL;
		fork->has.fe->number = node->number;
		return;
	}

	struct connfork *cfstack[MAXCONNR];
	int lengthstack[MAXCONNR];
	int cposstack[MAXCONNR];
	int depth = 1;
	cfstack[0] = firstfork;
	cposstack[0] = 0;
	lengthstack[0] = 0;
	struct connfork *curfork;
	int cpos;	// Position in conns.
	int length=0;	// The number of connections in the set we're currently looking at.
	unsigned short subsetmask = 0; //	For each conn is super, does it exist in conns?
	struct connfork *superfork = NULL;

	while(depth)
	{
		curfork = cfstack[--depth];
		cpos = cposstack[depth];
		if(!superfork)
		{
			length = lengthstack[depth];
			subsetmask &= (1 << length) - 1;
		}
		while(1)		// While not in a dead end.
		{
			if(curfork->conn < conns[cpos])
			{
				// Our set does not have this connection.
				// Test both ways, if needed. 'has' exists always if isend is not set.
				if(curfork->hasnot)	// Stack it, possibly.
				{
					if(!curfork->isend)	// Two different ways.
					{
						cfstack[depth] = curfork->hasnot;
						if(!superfork)
						{
							lengthstack[depth] = length;
							subsetmask &= ~(1 << length++);
						}
						cposstack[depth++] = cpos;
						curfork = curfork->has.cf;
					}
					else
					{
						curfork = curfork->hasnot;
						// curfork->conn is not in the sets we're considering. Don't change the mask.
					}
				}
				else if(!curfork->isend)	// Only do non-endnodes.
				{
					curfork = curfork->has.cf;
					if(!superfork) subsetmask &= ~(1 << length++);
				}
				else
				{
					break;
				}
			}
			else if(curfork->conn == conns[cpos])
			{
				cpos++;
				if(cpos == len)	// Every connection has been matched!
				{
					if(superfork) return;	// There are several supersets!
					subsetmask |= (1 << length++);
					while(!curfork->isend)	// Not an end node
					{
						curfork = curfork->has.cf;
						if(curfork->hasnot) return;	// There are several supersets!
					}
					if(curfork->has.fe->number & 0x80000000) return;	// Degenerate.
					superfork = curfork;		// This might be something good.
					break;	// We've come to an end here.
				}
				else
				{
					if(!curfork->isend)	// Ignore end nodes that don't have all our conns.
					{
						curfork = curfork->has.cf;		// Walk deeper.
						if(!superfork) subsetmask |= (1 << length++);
					}
					else break;
				}
			}
			else
			{
				break;		// Dead end.
			}
		}
	}

	if(!superfork)	// No supersets found. Let's add this stuff, then.
	{
		cpos = 0;
		curfork = firstfork;
		struct connfork *newfork, *oldfork = NULL;
		while(cpos < len)
		{
			if(curfork->conn < conns[cpos])	// We don't have this one.
			{
				if(curfork->hasnot)
				{
					oldfork = curfork;
					curfork = curfork->hasnot;
				}
				else	// Add a 'has not' thingy and a bunch of stuff.
				{
					newfork = (struct connfork *)mymalloc(sizeof(struct connfork));
					newfork->conn = conns[cpos++];
					curfork->hasnot = newfork;
					curfork = newfork;
					while(cpos < len)
					{
						newfork = (struct connfork *)mymalloc(sizeof(struct connfork));
						newfork->conn = conns[cpos++];
						curfork->has.cf = newfork;
						curfork = newfork;
					}
					curfork->isend = 1;
					curfork->has.fe = (struct forkend *)mymalloc(sizeof(struct forkend));
					curfork->has.fe->subsets = NULL;
					curfork->has.fe->number = node->number;
					return;
				}
			}
			else if(curfork->conn == conns[cpos])	// We have this one.
			{
				oldfork = curfork;
				curfork = curfork->has.cf;
				cpos++;
			}
			else	// We have to insert our tree branch between oldfork and curfork.
			{
				newfork = (struct connfork *)mymalloc(sizeof(struct connfork));
				newfork->conn = conns[cpos++];
				newfork->hasnot = curfork;
				if(oldfork)
				{
					// What way were we going?
					if(oldfork->has.cf == curfork)
					{
						oldfork->has.cf = newfork;
					}
					else
					{
						oldfork->hasnot = newfork;
					}
				}
				else
				{
					firstfork = newfork;
				}
				curfork = newfork;
				while(cpos < len)
				{
					newfork = (struct connfork *)mymalloc(sizeof(struct connfork));
					newfork->conn = conns[cpos++];
					curfork->has.cf = newfork;
					curfork = newfork;
				}
				curfork->isend = 1;
				curfork->has.fe = (struct forkend *)mymalloc(sizeof(struct forkend));
				curfork->has.fe->subsets = NULL;
				curfork->has.fe->number = node->number;
				return;
			}
		}
	}
	else		// Really simple case. I need it after the 150 lines above...
	{
		// This is a subset of another set. Redundant?
		struct forkend *fe = superfork->has.fe;
		if(!fe->subsets)
		{
			fe->subsets = (unsigned short *)malloc(sizeof(unsigned short) * 2);
			fe->subsets[0] = 1;
			fe->subsets[1] = ~subsetmask;
			return;
		}
		else
		{
			for(int i = fe->subsets[0]; i > 0; i--)
			{
				if(!(subsetmask & fe->subsets[i])) return;
			}
		}
		unsigned short *fes = fe->subsets;
		if(fes[0] == 65535)
		{
			printf("Internal error: more than 0xffff subsets for one set!\n");
			exit(1);
		}
		fe->subsets = (unsigned short *)malloc(sizeof(unsigned short) * (fes[0] + 2));
		memcpy(&fe->subsets[1], &fes[1], sizeof(unsigned short) * fes[0]);
		fe->subsets[0] = fes[0] + 1;
		fe->subsets[fe->subsets[0]] = ~subsetmask;
		free(fes);
	}
}

void do_the_subset_thing(int nodes)
{
	struct node_d *node, *oldnode;
	int tested = 0, degen = 0;

	if(!nodes) return;
	node = d_arr;
	oldnode = node;

	for(int i = 1; i < nodes; i++)
	{
		node++;
		// Do they differ?
		if(memcmp(node->conns, oldnode->conns, oldnode->connlen))
		{
			try_add_subset(oldnode);
			if(oldnode->number & 0x80000000) degen++;
			oldnode = node;
			tested++;
		}
		else oldnode->number |= 0x80000000;
	}
	try_add_subset(oldnode);
	if(oldnode->number & 0x80000000) degen++;
	tested++;
	printf("Examined %d (%d degen) / %d sets.\n", tested, degen, nodes);
}

void sort_temp_files(void)
{
	int fnum;
	for(fnum = 0; fnum < nr_tempfiles; fnum++)
	{
		// Read the temp file.
		if(fread(d_arr, sizeof(struct node_d) * tempfileentries[fnum], 1, tempfiles[fnum]) != 1)
		{
			perror("Read error on temp file");
			exit(1);
		}
		quick_sort(tempfileentries[fnum]);
		do_the_subset_thing(tempfileentries[fnum]);
	}
}

struct forkend *nextforkend(int cont, unsigned char *conns, int *length, int *common)
{
	static struct connfork *cfstack[MAXCONNR];
	static int lenstack[MAXCONNR];
	static int depth;
	struct connfork *curfork;
	int len;

	if(!cont)
	{
		if(!firstfork) return NULL;
		depth = 1;
		lenstack[0] = 0;
		cfstack[0] = firstfork;
	}
	if(!depth) return NULL;
	curfork = cfstack[--depth];
	len = lenstack[depth];
	*common = len;
	while(1)
	{
		if(curfork->hasnot)
		{
			lenstack[depth] = len;
			cfstack[depth++] = curfork->hasnot;
		}
		conns[len++] = curfork->conn;
		if(curfork->isend)
		{
			*length = len;
			return curfork->has.fe;
		}
		curfork = curfork->has.cf;
	}
	return NULL;	// Silly compiler.
}

void write_unpacked_output(void)
{
	struct node_d outnode, subnode;
	int len, common;

	memset(&outnode, 0, sizeof(struct node_d));
	memset(&subnode, 0, sizeof(struct node_d));
	outnode.wilds = 0;
	subnode.number = 0x80000000;
	subnode.wilds = 0;

	for(struct forkend *fe = nextforkend(0, outnode.conns, &len, &common); fe;
		fe = nextforkend(1, outnode.conns, &len, &common))
	{
		connlencount[0][len]++;

		outnode.connlen = len;
		outnode.number = fe->number;
		if(fe->number & 0x80000000)
			connlencount[1][len]++;
		if(len != MAXNRCONNS) memset(&outnode.conns[len], 0, MAXNRCONNS - len);

		fwrite(&outnode, sizeof(struct node_d), 1, outfile);

		if(fe->subsets)
		{
//printf("Node   ");
//for(unsigned int x=0;x<sizeof(struct node_d);x++) printf("%02x ",((unsigned char *)&outnode)[x]);
//printf("\n");
			if(len != MAXNRCONNS) memset(&subnode.conns[len], 0, MAXNRCONNS - len);
			for(int i = fe->subsets[0]; i > 0; i--)
			{
				int l = 0, l2 = len;
				for(int p = 0; p < len; p++)
				{
					if(fe->subsets[i] & (1 << p))
						subnode.conns[--l2] = 0;
					else
						subnode.conns[l++] = outnode.conns[p];
				}
				subnode.connlen = l;

//printf("Subset ");
//for(unsigned int x=0;x<sizeof(struct node_d);x++) printf("%02x ",((unsigned char *)&subnode)[x]);
//printf("\n");
				fwrite(&subnode, sizeof(struct node_d), 1, outfile);
			}
			connlencount[2][len] += fe->subsets[0];
		}
	}
}

void close_and_delete()
{
	char namebuf[40];
	int curfile;

	for(curfile = 0; curfile < nr_tempfiles; curfile++)
	{
		sprintf(namebuf, "tmp_%d.out", curfile);
		fclose(tempfiles[curfile]);
		remove(namebuf);
	}
	remove(TEMPINFILENAME);
	fclose(outfile);
}

void do_sort(char *infn, char *outfn)
{
	alloc_and_clear(outfn);

	create_temp_input(infn);
	calc_and_open_tempfiles();
	distribute_temp_input();
	sort_temp_files();

	write_unpacked_output();

	int ts = 0, td = 0, ss = 0;
	for(int len = 0; len <= MAXNRCONNS; len++)
	{
		ts += connlencount[0][len];
		td += connlencount[1][len];
		ss += connlencount[2][len];
	}
	printf("Saved %d sets (%d degen) with %d subsets.\n", ts, td, ss);
	for(int len = 0; len <= MAXNRCONNS; len++)
	{
		if(connlencount[0][len])
			printf(" Length %2d: %d, %d, %d\n", len, connlencount[0][len],
				connlencount[1][len], connlencount[2][len]);
	}
	close_and_delete();
}


int main(int argc, char **argv)
{
	char fnamebuf[2][32];

	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);

	printf("[sort_p2_input]\n");
	printf("CHAINLENGTH,MAXNRCONNS,MAXCONNR: %d,%d,%d\n", CHAINLENGTH, MAXNRCONNS, MAXCONNR);
	printf("sizeof(struct node_d),MAXFILESIZE,MAXFILEENTRIES: %zd,%d,%zd\n",
		sizeof(struct node_d), MAXFILESIZE, MAXFILEENTRIES);

	if(argc == 2) do_sort(argv[1], argv[2]);
	else if(argc < 2)
	{
//		sprintf(fnamebuf[0], "n_%02d.conn", CHAINLENGTH);
		sprintf(fnamebuf[0], "sab5n%02d.out", CHAINLENGTH);
		sprintf(fnamebuf[1], "n_%02d.conn2", CHAINLENGTH);
		do_sort(fnamebuf[0], fnamebuf[1]);
	}
	else
	{
		printf("Syntax: sort_p2_input [input_file output_file]\n");
	}

	fflush(stdout);
//	system("ps v");

	return 0;
}
