#ifndef BASIC_STATE_H
#define BASIC_STATE_H

#include <vector>
#include <algorithm>
#include "gsl_rng.h"

struct basic_state
{
	basic_state(int s): _state(s)
	{
	}
	unsigned int size() const
	{
		return _state.size();
	}
	void swap(basic_state &s2)
	{
		_state.swap(s2._state);
	}
	statetype &operator[](unsigned int pos)
	{
		return _state[pos];
	}
	const statetype &operator[](unsigned int pos) const
	{
		return _state[pos];
	}
	bool operator<(const basic_state &s2) const
	{
		return _state < s2._state;
	}
	bool operator==(const basic_state &s2) const
	{
		return _state == s2._state;
	}
	int hammingDistance(const basic_state &s2)
	{
		int si = size(), h = 0;
		for(int i = 0; i < si; i++)
		{
			if((*this)[i] != s2[i])
				h++;
		}
		return h;
	}
	void randomize()
	{
		const int s = size();
		for(int i = 0; i < s; i++)
			_state[i] = gsl_rng_uniform_int(rng, 2);
	}

	void modifyStateByHamming(int dist)
	{
		int si = size();
		if(si != (int)_htmp.size())
		{
			_htmp.resize(si);
			for(int i = 0; i < si; i++)
				_htmp[i] = i;
		}
		if(dist > si >> 1)
		{
			dist = si - dist;
			for(int i = 0; i < si; i++)
				_state[i] ^= 1;
		}
		for(int i = 0; i < dist; i++)
		{
			int lp = si - i;
			int pos = gsl_rng_uniform_int(rng, lp);
			_state[_htmp[pos]] ^= 1;
			if(pos != lp - 1)
				std::swap(_htmp[pos], _htmp[lp - 1]);
		}
	}


	static std::vector<nodeid> _htmp;
	std::vector<statetype> _state;
};



#endif
