#ifndef BOOL_KNET_H
#define BOOL_KNET_H

#include <vector>
#include <iostream>
#include <fstream>
#include <map>
#include "gsl_rng.h"

extern gsl_rng *rng;

typedef short nodeid;
typedef std::vector<unsigned int> smallstate;
typedef char statetype;

typedef std::vector<nodeid> nodevec;
template<class rule_T>
struct knode_basic
{
	nodevec inputs;
	rule_T rule;
	nodeid extra;
};

template<class rule_T>
struct knet_basic
{
	typedef std::vector<knode_basic<rule_T> > type;
};

#include "rule_canal2.h"

typedef rule_canal2 rule_t;

typedef rule_t::state_t state;
typedef knode_basic<rule_t> knode;
typedef knet_basic<rule_t>::type knet;

typedef short cyclen_t;
struct cycle
{
	cycle(int n):s(n){}
	cyclen_t len;
	smallstate s;
};
inline bool operator<(const cycle &c1, const cycle &c2)
{
	return (c1.len < c2.len) || (c1.len == c2.len && c1.s < c2.s);
}

class Knet
{
	struct edgeftype
	{
		int n1, n2, ftype, inv;
		edgeftype(int ln1, int ln2, int lftype, int linv):
			n1(ln1), n2(ln2), ftype(lftype), inv(linv)
		{
		}
	};

	struct efstats
	{
		efstats(): v(5) {}
		std::vector<int> v;
	};
	Knet() {}

	// Used by killNonRegulatingNodes
	static void killNonRegRecursion(knet& net, int n);
	// Turns a state into a space-saving smallstate
	static void statecompress(state &st, smallstate &sst);
	// Helper for findcycle
	static void findlowest(knet &net, state &s1, state &s2, state &ssave, cyclen_t len);
	// Removes (recursively) all nodes that do not have any outputs. Call
	// shrinkNet() afterwards.
	static void killNonRegulatingNodes(knet& net);
	// Removes nodes with invalid rules (i.e. nodes flagged as redundant)
	static void shrinkNet(const knet& net, knet& net2);

  public:
	// Makes a copy of a net and randomizes its rules (keeping the structure)
	static void prepareNet(const knet &rawnet, knet& net);
	// Prepare a net in place, with no node removal
	static void prepareNetNoshrink(knet& net);
	// Loads a network from an istream. Each line should be on the format
	// "outnode innode", where innode and outnode are strings. The mapping
	// from node name to number will be stored in nodenames.
	static void loadNet(std::istream &in, knet &net,
		std::vector<std::string> &nodenames);
	// Prints a state to cout
	static void printstate(state &st);
	// Starts from state s1 (which is randomized if rand is true), finds
	// a cycle, and stores a state identifying the cycle in ssave. The
	// length is stored in len, and 0 indicates failure to find a cycle in
	// turnlimits time steps. s2 is a temporary state. All states must be
	// constructed with the network size.
	static void findcycle(knet &net, state &s1, state &s2, state &ssave, cyclen_t &len,
		bool rand = true, const int turnlimit = 8192);
	// Print headers for statisticsOnCycles()
	static void statisticsOnCyclesHeader(const int starts, const int nets,
		std::ostream &out, int maxlen);
	// Produces statistics on the number of cycles of different lengths
	static void statisticsOnCycles(const knet &rawnet, const int starts,
		const int nets, std::ostream &out, std::ostream &err, bool synch,
		int maxlen, bool printheader = true);
	// Vector (non-print) version of the above (with progress output)
	static void gatherStatisticsOnCycles(const knet &rawnet, const int starts,
		const int nets, std::ostream &err, bool synch,
		int maxlen, std::vector<double> &means);
	// Examines how well states of nodes correlate between fixed points
	static void nodeCorrInFixedpoints(const knet &rawnet, const int starts,
		const int nets, std::ostream &out, std::ostream &err, bool synch,
		std::ostream &img);
	// Makes a derrida plot
	static void derridaPlot(const knet &rawnet, const int starts, const int nets,
		std::ostream &out, std::ostream &err, int presteps, int derrsteps);
	// derrida plot, adding to the vectors output/outputsqr
	static void derridaPlotStats(const knet &rawnet, const int starts,
		const int nets, std::ostream &err, int presteps, int derrsteps,
		std::vector<double> &output, std::vector<double> &outputsql);
	// Looks at the hamming distance distribution for one initial hamming
	// distance. In a derrida plot, the mean of this distribution is
	// plotted vs initial H.
	static void hammingDistrPlot(const knet &rawnet, const int starts, const int nets,
		std::ostream &out, std::ostream &err, int presteps, int derrsteps,
		int hamming);
	// Makes a plot of the time development of H between two random states
	// starting from distance 'hamming'.
	static void timeCoursePlot(const knet &rawnet, const int starts, const int nets,
		std::ostream &out, std::ostream &err, int presteps, int gathersteps,
		int hamming);
	// Similar, but starting from a fixed point, weighted by the fixed point's
	// attractor basin size.
	static void timeCoursePlotFromFP(const knet &rawnet, const int starts,
		const int nets, std::ostream &out, std::ostream &err, int gathersteps,
		int hamming);
	// Examines the probability of converging to a fixed point F after
	// being perturbed from it, as a function of the perturbation size.
	static void probabilityForFP(const knet &rawnet, const int starts,
		const int nets, std::ostream &out, std::ostream &err);
	// Copies the net, freezes trivial forcing structures, and colors
	// fixed nodes after their frozen state. 'colors' maps node # to color.
	static void colorNetwork(const knet &orignet, std::map<int, int> &colors);
	// Calculates for each connection whether it's forcing, and if so
	// how. Returns the fraction of canalyzing connections.
	static double calculateForcingness(knet &net, std::vector<edgeftype> &edges);
	// Outputs the mean degree of forcingness for all connections,
	// and returns the grand mean. The output is meant to be processed
	// by the dottify_color script.
	static double averageForcing(const knet &rawnet, std::ostream &out,
		std::ostream &err, int nets, const std::vector<std::string> &nodenames);
	// Outputs info about frozen nodes and forcing structures in a
	// single network.
	static double examineForcing(const knet &rawnet, std::ostream &out,
		std::ostream &err, const std::vector<std::string> &nodenames);

};


#endif
