/**
 * Filename     : baseCompartmentNeighborhood.cc
 * Description  : A base class describing variable updates
 * Author(s)    : Henrik Jonsson (henrik@thep.lu.se)
 * Created      : October 2003
 * Revision     : $Id: baseCompartmentNeighborhood.cc,v 1.7 2006/03/18 00:05:14 henrik Exp $
 */
#include<vector>

#include"baseCompartmentNeighborhood.h"
#include"compartmentNeighborhood.h"

BaseCompartmentNeighborhood::~BaseCompartmentNeighborhood(){}


//!Factory creator, all creation should be mapped onto this one 
/*! Given the idValue a compartmentNeighborhood of the defined type is returned
(using new Class).*/
BaseCompartmentNeighborhood *
BaseCompartmentNeighborhood::
createCompartmentNeighborhood(std::vector<double> &paraValue,
			      std::vector< std::vector<size_t> > &indValue, 
			      std::string idValue ) {
  
  // Distance rules
  //(compartmentNeighborhood.h,compartmentNeighborhood.cc)
  //////////////////////////////////////////////////////////////////////
  if( idValue=="neighborhoodDistanceSphere" ) 
    return new NeighborhoodDistanceSphere(paraValue,indValue);
  //Default, if nothing found
  //////////////////////////////////////////////////////////////////////
  else {
    std::cerr << "\nBaseCompartmentNeighborhood::"
	      << "createCompartmentNeighborhood() "
	      << "WARNING: CompartmentNeighborhoodtype " << idValue 
	      << " not known, no compartmentNeighborhood created.\n\7";
    exit(-1);
  }
}

//!This creator reads from an open file and then call the main creator
BaseCompartmentNeighborhood * 
BaseCompartmentNeighborhood::
createCompartmentNeighborhood(std::istream &IN ) {
  
 
  std::string idVal;
  size_t pNum,levelNum;
  IN >> idVal;
  
  IN >> pNum;
  IN >> levelNum;
  std::vector<size_t> varIndexNum( levelNum );
  for( size_t i=0 ; i<levelNum ; i++ )
    IN >> varIndexNum[i];
  
  std::vector<double> pVal( pNum );
  for( size_t i=0 ; i<pNum ; i++ )
    IN >> pVal[i];
  
  std::vector< std::vector<size_t> > varIndexVal( levelNum );
  for( size_t i=0 ; i<levelNum ; i++ )
    varIndexVal[i].resize( varIndexNum[i] );
  
  for( size_t i=0 ; i<levelNum ; i++ )
    for( size_t j=0 ; j<varIndexNum[i] ; j++ )
      IN >> varIndexVal[i][j];
  
 
  return createCompartmentNeighborhood(pVal,varIndexVal,idVal);
}

unsigned int BaseCompartmentNeighborhood::
create(Organism &O,std::vector< std::vector<double> > &y,double t ) {
  std::cerr << "BaseCompartmentNeighborhood::create() should not be used. "
	    << "Should always be mapped onto one of the real types.\n";
  exit(0);
}  

unsigned int BaseCompartmentNeighborhood::
update(Organism &O,std::vector< std::vector<double> > &y,double t ) {
  std::cerr << "BaseCompartmentNeighborhood::update() should not be used. "
	    << "Should always be mapped onto one of the real types.\n";
  exit(0);
}  
